/*
 * This driver forms two instances. Instance 0 works as the input of
 * a pipe (the writers side), instance 1 is the output (the readers side).
 * The pipe holds one single record of size IOBUFLEN. If the pipe is empty
 * readers block. If it is full writers block.
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>

#include "mypipe.h"

#include <sys/ddi.h>
#include <sys/sunddi.h>

#define	 MYPIPE_NAME	"mypipe"

/* 
 * The two instances has one minor number each.
 */
#define	MYPIPE_DEV_TO_INST(dev)			(getminor(dev))
#define	MYPIPE_INST_TO_DEV(instance)		(instance)
#define MYPIPE_INSTANCES 2

static char mypipe_buf[IOBUFLEN];
static kmutex_t	mypipe_mutex;
static kcondvar_t mypipe_full;
static kcondvar_t mypipe_empty;
static unsigned mypipe_is_full = 0;
static unsigned mypipe_readers = 0;
static unsigned mypipe_writers = 0;

static dev_info_t dip_vect[MYPIPE_INSTANCES];

static int mypipe_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp);
static int mypipe_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int mypipe_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int mypipe_open(dev_t *devp, int flag, int otyp, cred_t *cred);
static int mypipe_close(dev_t dev, int flag, int otyp, cred_t *cred);
static int mypipe_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int mypipe_write(dev_t dev, struct uio *uiop, cred_t *credp);

static struct cb_ops mypipe_cb_ops = {
	mypipe_open,		/* open */
	mypipe_close,		/* close */
	nodev,			/* strategy */
	nodev,			/* print */
	nodev,			/* dump */
	mypipe_read,		/* read */
	mypipe_write,		/* write */
	nodev,			/* ioctl */
	nodev,			/* devmap */
	nodev,			/* mmap */
	nodev,			/* segmap */
	nochpoll,		/* poll */
	ddi_prop_op,		/* cb_prop_op */
	NULL,			/* not a stream */
	D_NEW | D_MP | D_64BIT,
	CB_REV,
	nodev,			/* aread */
	nodev			/* awrite */
};


static struct dev_ops mypipe_ops = {
	DEVO_REV,		/* devo_rev     */
	0,			/* refcnt       */
	mypipe_getinfo,		/* info         */
	nulldev,		/* identify     */
	nulldev,		/* probe        */
	mypipe_attach,		/* attach       */
	mypipe_detach,		/* detach       */
	nodev,			/* reset        */
	&mypipe_cb_ops,
	(struct bus_ops *)NULL,
	nulldev			/* no mypipe_power() */
};

static struct modldrv md = {
	&mod_driverops, 	/* Type of module. This is a driver */
	"mypipe driver v0.1", 	/* Name of the module */
	&mypipe_ops,
};

static struct modlinkage ml = {
	MODREV_1,
	&md,
	NULL
};


static int mypipe_pflag = 1;

int
_init(void)
{
	if (mypipe_pflag) 
		cmn_err(CE_CONT, "mypipe: _init()");

	mutex_init(&mypipe_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&mypipe_empty, NULL, CV_DRIVER, NULL);
	cv_init(&mypipe_full, NULL, CV_DRIVER, NULL);

	return mod_install(&ml);
}

int
_info(struct modinfo *modinfop)
{
	if (mypipe_pflag) 
		cmn_err(CE_CONT, "mypipe: _info()");

	return mod_info(&ml, modinfop);
}

int
_fini(void)
{
	int	error;

	if (mypipe_pflag) 
		cmn_err(CE_CONT, "mypipe: _fini()");

	error = mod_remove(&ml);
	if (error)
		return error;

	mutex_destroy(&mypipe_mutex);
	cv_destroy(&mypipe_empty);
	cv_destroy(&mypipe_full);

	return 0;
}

static int
mypipe_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp)
{
	int instance;

	*resultp = NULL;
	instance = MYPIPE_DEV_TO_INST((dev_t)arg);
	if (instance < 0 || instance >= MYPIPE_INSTANCES)
		return DDI_FAILURE;

	switch (cmd) {
	    case DDI_INFO_DEVT2DEVINFO:
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
				"mypipe: mypipe_getinfo(DDI_INFO_DEVT2DEVINFO) i%d",
			    	instance);
		*resultp = dip_vect[instance];
		return DDI_SUCCESS;
	    case DDI_INFO_DEVT2INSTANCE: /* gerufen bei deferred attach */
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
			    "mypipe: mypipe_getinfo(DDI_INFO_DEVT2INSTANCE) i%d",
			    	instance);
		*resultp = (void *)MYPIPE_INST_TO_DEV(instance);
		return DDI_SUCCESS;
	    default:
		return DDI_FAILURE;
	}
}

static int
mypipe_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int	instance;

	instance = ddi_get_instance(dip);

	switch (cmd) {
	    case DDI_ATTACH:
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
				"mypipe: mypipe_attach(DDI_ATTACH), i%d", instance);
		dip_vect[instance] = dip;

		if (mypipe_pflag) 
			cmn_err(CE_CONT, "mypipe: mypipe_attach, %s%d,", 
				MYPIPE_NAME, instance);

		if (ddi_create_minor_node(dip, instance ? "out" : "in", 
				S_IFCHR, MYPIPE_INST_TO_DEV(instance), 
				DDI_PSEUDO, 0) != DDI_SUCCESS) {
			cmn_err(CE_NOTE,
				"%s%d: attach: could not add character node.",
				MYPIPE_NAME, instance);
			ddi_remove_minor_node(dip, NULL);
			return (DDI_FAILURE);
		}

		ddi_report_dev(dip);
		break;

	    default:
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
				"mypipe: mypipe_attach(%d), i%d", cmd, instance);
		return (DDI_FAILURE);
	}

	return (DDI_SUCCESS);
}

static int
mypipe_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int		instance;

	instance = ddi_get_instance(dip);

	switch (cmd) {
	case DDI_DETACH:
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
				"mypipe: mypipe_detach(DDI_DETACH), %s%d",
				MYPIPE_NAME, instance);

		ddi_remove_minor_node(dip, NULL);

		return (DDI_SUCCESS);
	default:
		if (mypipe_pflag) 
			cmn_err(CE_CONT, 
				"mypipe: mypipe_detach(%d), %s%d", 
					cmd, MYPIPE_NAME, instance);
		return (DDI_FAILURE);
	}
}

static int
mypipe_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
	int instance, openval;

	if (otyp != OTYP_CHR)
		return (EINVAL);

	instance = MYPIPE_DEV_TO_INST(*devp);
	
	if (mypipe_pflag) 
		cmn_err(CE_CONT, "mypipe: mypipe_open(0x%x) i%d", 
			flag, instance);

	if (instance == 0) {
		if ((flag & (FREAD|FWRITE)) != FWRITE)
			return EINVAL;

		mutex_enter(&mypipe_mutex);
		++mypipe_writers;
		mutex_exit(&mypipe_mutex);
	}
	else if (instance == 1) {
		if ((flag & (FREAD|FWRITE)) != FREAD)
			return EINVAL;

		mutex_enter(&mypipe_mutex);
		++mypipe_readers;
		mutex_exit(&mypipe_mutex);
	}
	else {
		cmn_err(CE_CONT, "mypipe: invalid instance %d", instance);
		return ENXIO;
	}

	return (0);
}

static int
mypipe_close(dev_t dev, int flag, int otyp, cred_t *cred)
{
	int	instance;

	instance = MYPIPE_DEV_TO_INST(dev);
	if (mypipe_pflag) 
		cmn_err(CE_CONT, "mypipe: mypipe_close i%d", instance);

	if (instance == 0) {
		mutex_enter(&mypipe_mutex);
		mypipe_writers = 0;
		mutex_exit(&mypipe_mutex);
	}
	else if (instance == 1) {
		mutex_enter(&mypipe_mutex);
		mypipe_readers = 0;
		mutex_exit(&mypipe_mutex);
	}
	else
		return ENXIO;

	return (0);
}

static int
mypipe_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
	int instance;

	instance = MYPIPE_DEV_TO_INST(dev);
	if (instance != 1)
		return ENXIO;

	if (uiop->uio_resid != IOBUFLEN)
		return EINVAL;

	mutex_enter(&mypipe_mutex);

	while (!mypipe_is_full) {
		if (!cv_wait_sig(&mypipe_full, &mypipe_mutex)) {
			mutex_exit(&mypipe_mutex);
			return EINTR;
		}
	}

	uiomove(mypipe_buf, uiop->uio_resid, UIO_READ, uiop);
	mypipe_is_full = 0;

	mutex_exit(&mypipe_mutex);

	/* wake up users waiting for state == IDLE */
	cv_signal(&mypipe_empty); 

	return (0);
}

static int
mypipe_write(dev_t dev, struct uio *uiop, cred_t *credp)
{
	int instance;

	instance = MYPIPE_DEV_TO_INST(dev);
	if (instance != 0)
		return ENXIO;

	if (uiop->uio_resid != IOBUFLEN)
		return EINVAL;

	mutex_enter(&mypipe_mutex);

	while (mypipe_is_full) {
		if (!cv_wait_sig(&mypipe_empty, &mypipe_mutex)) {
			mutex_exit(&mypipe_mutex);
			return EINTR;
		}
	}

	uiomove(mypipe_buf, uiop->uio_resid, UIO_WRITE, uiop);
	mypipe_is_full = 1;

	mutex_exit(&mypipe_mutex);

	/* wake up users waiting for state == IDLE */
	cv_signal(&mypipe_full); 

	return (0);
}
