#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>

#include "loop.h"
#include <sys/ddi.h>
#include <sys/sunddi.h>

static int loop_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp);
static int loop_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int loop_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int loop_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int loop_close(queue_t *q, int oflag, cred_t *credp);
static int loop_wput(queue_t *q, mblk_t *mp);
static int loop_wsrv(queue_t *q);
static int loop_rsrv(queue_t *q);

static void comment(queue_t *q, mblk_t **mp);
static void print_msg(char *s, mblk_t *mp);
static void do_ioctl(queue_t *q, mblk_t *mp);

static struct module_info rminfo = {0x123, "loop", 0, INFPSZ, STRLOW, STRHIGH};

static struct module_info wminfo = {0x123, "loop", 0, INFPSZ, STRLOW, STRHIGH};

static struct qinit rinit = {
	NULL, loop_rsrv, loop_open, loop_close, NULL, &rminfo, NULL,
};

static struct qinit winit = {
	loop_wput, loop_wsrv, NULL, NULL, NULL, &wminfo, NULL,
};

struct streamtab loop_strtab = {
	&rinit, &winit, NULL, NULL, 
};

static struct cb_ops loop_cb_ops = {
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nochpoll,		/* cb_chpoll 		*/
	ddi_prop_op,		/* cb_prop_op 		*/
	&loop_strtab,		/* cb_str, streamtab 	*/
	D_NEW|D_MP|D_MTPERMOD,	/* cb_flag 		*/
	CB_REV,			/* cb_rev 		*/
	nodev,
	nodev
};

static struct dev_ops loop_ops = {
	DEVO_REV,		/* devo_rev		*/
	0,			/* ref count 		*/
	loop_getinfo,		/* devo_getinfo		*/
	nulldev,		/* devo_identify	*/
	nulldev,		/* devo_probe		*/
	loop_attach,		/* devo_attach		*/
	loop_detach,		/* devo_detach		*/
	nodev,			/* devo_reset		*/
	&loop_cb_ops,		/* devo_cb_ops		*/
	NULL,			/* devo_bus_ops		*/
	nulldev			/* devo_power		*/
};

static struct modldrv md = {
	&mod_driverops, 	/* Type of module. This is a driver */
	"loop driver v0.1", 	/* Name of the module */
	&loop_ops,
};

static struct modlinkage ml = {
	MODREV_1,
	&md,
	NULL
};

/* device to instance mapping: every instance has one minor (= 0) */
#define  LOOP_DEV_TO_INST(dev)                   (getminor(dev))

static int loop_pflag = 1;

/* private loop instance structure */
struct loop_state {
	int		flags;		/* values see below		*/
	dev_info_t	*dip;		/* device info pointer		*/
	queue_t 	*q;		/* pointer to read queue 	*/
	struct loop_statistics stat;	/* statistics data		*/
};

#define LOOP_MAX_INST 100
static struct loop_state loop_state[LOOP_MAX_INST]; /* indexed by instance */

static int max_inst = 0; /* instances from 0 ... max_inst-1 may be open */

/* flags values */
#define LOOP_INST_INUSE 	0x01

int
_init(void)
{
	int e;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: _init()");

	return mod_install(&ml);
}

int
_info(struct modinfo *modinfop)
{
	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: _info()");

	return mod_info(&ml, modinfop);
}

int
_fini(void)
{
	int e;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: _fini()");

	return(mod_remove(&ml));
}

static int
loop_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp)
{
	int instance;

	*resultp = NULL;
	instance = LOOP_DEV_TO_INST((dev_t)arg);

	switch (cmd) {
	    case DDI_INFO_DEVT2DEVINFO:
		if (loop_pflag) 
			cmn_err(CE_CONT, 
				"loop: loop_getinfo(DDI_INFO_DEVT2DEVINFO) i%d",
			    	instance);

		if (instance < 0 || instance >= LOOP_MAX_INST ||
			!(loop_state[instance].flags & LOOP_INST_INUSE) || 
						!loop_state[instance].dip)
			return DDI_FAILURE;

		*resultp = loop_state[instance].dip;
		return DDI_SUCCESS;

	    case DDI_INFO_DEVT2INSTANCE: /* gerufen bei deferred attach */
		if (loop_pflag) 
			cmn_err(CE_CONT, 
			"loop: loop_getinfo(DDI_INFO_DEVT2INSTANCE) i%d",
			    	instance);
		*resultp = (void *)instance;
		return DDI_SUCCESS;

	    default:
		return DDI_FAILURE;
	}
}

static int
loop_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int	instance;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: attach()");

	switch (cmd) {
	    case DDI_ATTACH:
		instance = ddi_get_instance(dip);
		if (loop_pflag) 
			cmn_err(CE_CONT, 
				"loop: loop_attach(DDI_ATTACH), i%d", instance);

		if (ddi_create_minor_node(dip, "loop", S_IFCHR,
			ddi_get_instance(dip), DDI_PSEUDO, CLONE_DEV) 
							!= DDI_SUCCESS) {
				cmn_err(CE_NOTE,
				    "loop attach: could not create node %d",
				    instance);
				goto out;
		}

		ddi_report_dev(dip);
		return (DDI_SUCCESS);
	    default:
		instance = ddi_get_instance(dip);
		if (loop_pflag) 
			cmn_err(CE_CONT, 
				"loop: loop_attach(%d), i%d", cmd, instance);
		return (DDI_FAILURE);
	}
out:
	loop_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}

static int
loop_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int		instance;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: dettach()");

	switch (cmd) {
	    case DDI_DETACH:
		instance = ddi_get_instance(dip);
		if (loop_pflag) 
			cmn_err(CE_CONT, 
				"loop: loop_detach(DDI_DETACH), i%d", instance);
		ddi_remove_minor_node(dip, NULL);
		loop_state[instance].flags = 0;

		return (DDI_SUCCESS);

	    default:
		instance = ddi_get_instance(dip);
		if (loop_pflag) 
			cmn_err(CE_CONT, 
				"loop: loop_detach(%d), i%d", cmd, instance);
		return (DDI_FAILURE);
	}
}

static int 
loop_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	int instance;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: open()");
	
	if (q->q_ptr)
		return EBUSY;
	
	if (sflag != CLONEOPEN)
		return ENXIO;

	for (instance = 0; instance < max_inst; ++ instance)
		if (!(loop_state[instance].flags & LOOP_INST_INUSE))
			break;
	if (instance == max_inst)
		if (max_inst == LOOP_MAX_INST)
			return EMFILE;
		else
			++max_inst;
	*devp = makedevice(getmajor(*devp), instance);

	bzero(loop_state + instance, sizeof(struct loop_state));
	loop_state[instance].flags |= LOOP_INST_INUSE;
	loop_state[instance].q = q;
	WR(q)->q_ptr = q->q_ptr = (queue_t *)(loop_state + instance);

	qprocson(q);

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop open: instance %d%s",
			instance, sflag == CLONEOPEN ? " CLONE" : "");

	return 0;
}

static int 
loop_close(queue_t *q, int oflag, cred_t *credp)
{
	struct loop_state *ms;

	ms = (struct loop_state *)q->q_ptr;

	if (loop_pflag) 
		cmn_err(CE_CONT, "loop: close(), i %d", ms - loop_state);

	qprocsoff(q);

	ms->flags &= ~LOOP_INST_INUSE;
	ms->q = NULL;
	ms->dip = NULL;

	return 0;
}

static int 
loop_wput(queue_t *q, mblk_t *mp)
{
	struct loop_ctl *ctl;
	struct loop_state *ls = (struct loop_state *)q->q_ptr;
	struct loop_statistics *s;
	mblk_t *mp2;

	switch (mp->b_datap->db_type) {
	    case M_PCPROTO:
		/* high priority message: get status, non blocking */
		if (mp->b_wptr - mp->b_rptr != sizeof(struct loop_ctl) ||
		  !(mp2 = mp->b_cont) || 
		  mp2->b_wptr - mp2->b_rptr != sizeof(struct loop_statistics)){
			qreply(q, mp);
			break;
		}
		ctl = (struct loop_ctl *)mp->b_rptr;
		if (ctl->cmd != LOOP_C_STAT) {
			qreply(q, mp);
			break;
		}
		ctl->error = 0;
		s = (struct loop_statistics *)mp2->b_rptr;
		bcopy(&ls->stat, s, sizeof(struct loop_statistics));
		s->q_count = q->q_count;
		s->q_mblkcnt = q->q_mblkcnt;
		qreply(q, mp);
		break;

	    case M_PROTO:
	    case M_DATA:
		/* message is looped back upstream with comment added */
		comment(q, &mp);
		if (q->q_first || !canputnext(RD(q)))
			putq(q, mp);
		else
			qreply(q, mp);
		break;
	
	    case M_IOCTL:
		if (q->q_first)
			putq(q, mp);
		else
			do_ioctl(q, mp);
		break;

	    case M_FLUSH:
		/* canonical flush handling, read q is unused */
		if (*mp->b_rptr & FLUSHW)
			/* flush write q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		if (*mp->b_rptr & FLUSHR) {
			/* send request back upstream without FLUSHW */
			*mp->b_rptr &= ~FLUSHW;
			qreply(q, mp);
		}
		else
			freemsg(mp); /* free message */
		break;
	
	    default:
		/* message is looped back upstream, unchanged */
		if (q->q_first || !canputnext(RD(q)))
			putq(q, mp);
		else
			qreply(q, mp);
	    	break;
	}

	return 0;
}

static int 
loop_wsrv(queue_t *q)
{
	mblk_t *mp;

	while (mp = getq(q)) {

		switch (mp->b_datap->db_type) {
	    	    case M_IOCTL:
			do_ioctl(q, mp);
			break;

		    default:
			if (!canputnext(RD(q))) {
				putbq(q, mp);
				/* this q will not be scheduled any more 
			 	* until backenabled by loop_rsrv()
			 	*/
				return 0;
			}

			qreply(q, mp);
			break;
		}
	}

	return 0;
}

static int 
loop_rsrv(queue_t *q)
{
	if (loop_pflag) 
		cmn_err(CE_CONT, "loop_rsrv: backenabling write q");

	/* Called when q is backenabled. Backenable write q. */
	qenable(WR(q));

	return 0;
}

/*
 * Append "|loop %d|" with the driver's instance number to 
 * a message. The mblk_t might have to be copied. Therefore mp is 
 * of type mblk_t **.
 */
static void 
comment(queue_t *q, mblk_t **mp)
{
	struct loop_state *ls = (struct loop_state *)q->q_ptr;
	int instance = ls - loop_state;
	mblk_t *n, *prev, *h;
	char str[20];
	int l, l2;

	++ls->stat.num_msg;

	/* string to append to message */
	sprintf(str, "|loop %d|", instance);
	l = strlen(str);

	/* find last block in message */
	for (n = *mp, prev = NULL; n->b_cont; prev = n, n = n->b_cont)
		;

	/* if block n is referenced by others or has not enough space
	 * copy contents to a new and big enough block
	 */
	if (n->b_datap->db_ref > 1 || n->b_wptr + l >= n->b_datap->db_lim) {
		++ls->stat.num_allocb;

		l2 = n->b_wptr - n->b_rptr;
		h = allocb(l2 + l, BPRI_MED);
		if (!h) {
			++ls->stat.num_nomem;
			return; /* no memory, ignore it */
		}
		bcopy(n->b_rptr, h->b_wptr, l2);
		h->b_wptr += l2;
		if (prev)
			prev->b_cont = h;
		else
			*mp = h;
		freeb(n);
		n = h;
	}
	
	bcopy(str, n->b_wptr, l);
	n->b_wptr += l;
}

/*
 * print a message for debug purposes
 */
static void 
print_msg(char *s, mblk_t *mp)
{
	unsigned char b[80], *p, *p2;

	strncpy((char *)b, s, sizeof(b) - 1);
	b[sizeof(b) - 1] = 0;
	for (p = b + strlen((char *)b), p2 = mp->b_rptr; 
			p < b + sizeof(b) - 1 && p2 < mp->b_wptr; )
		*p++ = *p2++;
	*p = 0;

	printf("%s", b);
}

static void 
do_ioctl(queue_t *q, mblk_t *mp)
{
	struct loop_state *ls = (struct loop_state *)q->q_ptr;
	struct iocblk *ioc;
	struct loop_statistics *s;
	mblk_t *mp2;
	int err;

	ioc = (struct iocblk *) mp->b_rptr;
	if (ioc->ioc_count == TRANSPARENT) {
		err = EINVAL;
		goto ioc_nak;
	}
	switch (ioc->ioc_cmd) {
	    case LOOP_I_GETSTAT:
		if (ioc->ioc_count != sizeof(struct loop_statistics) ||
				!(mp2 = mp->b_cont) || 
		  		mp2->b_wptr - mp2->b_rptr != 
					sizeof(struct loop_statistics)) {
			err = EINVAL;
			goto ioc_nak;
		}
		s = (struct loop_statistics *)mp2->b_rptr;
		bcopy(&ls->stat, s, sizeof(struct loop_statistics));
		s->q_count = q->q_count;
		s->q_mblkcnt = q->q_mblkcnt;
		break;

	    default:
		err = EINVAL;
		goto ioc_nak;
	}

	/* send ACK */
	mp->b_datap->db_type = M_IOCACK;
	ioc->ioc_error = ioc->ioc_rval = 0;
	qreply(q, mp);

	return;

ioc_nak:
	mp->b_datap->db_type = M_IOCNAK;
	ioc->ioc_error = err;
	qreply(q, mp);
}

