#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>

#include "mux.h"
#include <sys/ddi.h>
#include <sys/sunddi.h>

static int mux_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp);
static int mux_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int mux_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int mux_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int mux_close(queue_t *q, int oflag, cred_t *credp);

static int mux_uwput(queue_t *q, mblk_t *mp);
static int mux_uwsrv(queue_t *q);
static int mux_lwsrv(queue_t *q);

static int mux_lrput(queue_t *q, mblk_t *mp);
static int mux_lrsrv(queue_t *q);
static int mux_ursrv(queue_t *q);

static void print_msg(char *s, mblk_t *mp);
static void do_ioctl(queue_t *q, mblk_t *mp);
static struct mux_ctl *is_mux_ctl(mblk_t *mp);
int route_downstream(queue_t *q, mblk_t *mp, int called_by_srv);
int route_upstream(queue_t *q, mblk_t *mp, int called_by_srv);

static struct module_info minfo = {0x123, "mux", 0, INFPSZ, STRLOW, STRHIGH};

static struct qinit uwinit = {
	mux_uwput, mux_uwsrv, NULL, NULL, NULL, &minfo, NULL,
};

static struct qinit urinit = {
	NULL, mux_ursrv, mux_open, mux_close, NULL, &minfo, NULL,
};

static struct qinit lwinit = {
	NULL, mux_lwsrv, NULL, NULL, NULL, &minfo, NULL,
};

static struct qinit lrinit = {
	mux_lrput, mux_lrsrv, NULL, NULL, NULL, &minfo, NULL,
};

struct streamtab mux_strtab = {
	&urinit, &uwinit, &lrinit, &lwinit, 
};

static struct cb_ops mux_cb_ops = {
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nochpoll,		/* cb_chpoll 		*/
	ddi_prop_op,		/* cb_prop_op 		*/
	&mux_strtab,		/* cb_str, streamtab 	*/
	D_NEW|D_MP|D_MTPERMOD,	/* cb_flag 		*/
	CB_REV,			/* cb_rev 		*/
	nodev,
	nodev
};

static struct dev_ops mux_ops = {
	DEVO_REV,		/* devo_rev		*/
	0,			/* ref count 		*/
	mux_getinfo,		/* devo_getinfo		*/
	nulldev,		/* devo_identify	*/
	nulldev,		/* devo_probe		*/
	mux_attach,		/* devo_attach		*/
	mux_detach,		/* devo_detach		*/
	nodev,			/* devo_reset		*/
	&mux_cb_ops,		/* devo_cb_ops		*/
	NULL,			/* devo_bus_ops		*/
	nulldev			/* devo_power		*/
};

static struct modldrv md = {
	&mod_driverops, 	/* Type of module. This is a driver */
	"mux driver v0.1", 	/* Name of the module */
	&mux_ops,
};

static struct modlinkage ml = {
	MODREV_1,
	&md,
	NULL
};

/* device to instance mapping: every instance has one minor (= 0) */
#define  MUX_DEV_TO_INST(dev)                   (getminor(dev))

static int mux_pflag = 1;

/* private mux instance structure */
struct mux_state {
	int		flags;		/* values see below		*/
	dev_info_t	*dip;		/* device info pointer		*/
	queue_t 	*q;		/* pointer to read queue 	*/
	struct mux_statistics stat;	/* statistics data		*/
};

#define MUX_MAX_INST 100
static struct mux_state mux_state[MUX_MAX_INST]; /* indexed by instance */

static int max_inst = 0; /* instances from 0 ... max_inst-1 may be open */

/* flags values */
#define MUX_INST_INUSE 		0x01

/* vector of lower streams managed */
struct mux_lower {
	int		flags;		/* see below			*/
	queue_t 	*q;		/* pointer to write queue 	*/
};

#define MAX_MUX_LOWER 100
static struct mux_lower mux_lower[MAX_MUX_LOWER];

/* Lower streams 0 ... max_lower-1 may be linked underneath mux. 
 * struct linkblk.l_index is an index into mux_lower[].
 */
static int max_lower = 0; 

/* flags values */
#define MUX_LOWER_INUSE 	0x01

int
_init(void)
{
	int e;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: _init()");

	e = mod_install(&ml);
	if (e)
		return e;

	return 0;
}

int
_info(struct modinfo *modinfop)
{
	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: _info()");

	return mod_info(&ml, modinfop);
}

int
_fini(void)
{
	int e;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: _fini()");
	
	return(mod_remove(&ml));
}

static int
mux_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp)
{
	int instance;

	*resultp = NULL;
	instance = MUX_DEV_TO_INST((dev_t)arg);

	switch (cmd) {
	    case DDI_INFO_DEVT2DEVINFO:
		if (mux_pflag) 
			cmn_err(CE_CONT, 
				"mux: mux_getinfo(DDI_INFO_DEVT2DEVINFO) i%d",
			    	instance);

		if (instance < 0 || instance >= MUX_MAX_INST ||
			!(mux_state[instance].flags & MUX_INST_INUSE) || 
						!mux_state[instance].dip)
			return DDI_FAILURE;

		*resultp = mux_state[instance].dip;
		return DDI_SUCCESS;

	    case DDI_INFO_DEVT2INSTANCE: /* gerufen bei deferred attach */
		if (mux_pflag) 
			cmn_err(CE_CONT, 
			"mux: mux_getinfo(DDI_INFO_DEVT2INSTANCE) i%d",
			    	instance);
		*resultp = (void *)instance;
		return DDI_SUCCESS;

	    default:
		return DDI_FAILURE;
	}
}

static int
mux_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int	instance;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: attach()");

	switch (cmd) {
	    case DDI_ATTACH:
		instance = ddi_get_instance(dip);
		if (mux_pflag) 
			cmn_err(CE_CONT, 
				"mux: mux_attach(DDI_ATTACH), i%d", instance);

		if (ddi_create_minor_node(dip, "mux", S_IFCHR,
			ddi_get_instance(dip), DDI_PSEUDO, CLONE_DEV) 
							!= DDI_SUCCESS) {
				cmn_err(CE_NOTE,
				    "mux attach: could not create node %d",
				    instance);
				goto out;
		}

		ddi_report_dev(dip);
		return (DDI_SUCCESS);
	    default:
		instance = ddi_get_instance(dip);
		if (mux_pflag) 
			cmn_err(CE_CONT, 
				"mux: mux_attach(%d), i%d", cmd, instance);
		return (DDI_FAILURE);
	}
out:
	mux_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}

static int
mux_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int		instance;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: dettach()");

	switch (cmd) {
	    case DDI_DETACH:
		instance = ddi_get_instance(dip);
		if (mux_pflag) 
			cmn_err(CE_CONT, 
				"mux: mux_detach(DDI_DETACH), i%d", instance);

		/* Without this permanently linked lower streams stay open
	 	 * without a reference to unlink them.
	 	 */
		if (max_lower > 0) {
			cmn_err(CE_WARN, "mux driver cannot detach");
			cmn_err(CE_CONT, "first unlink all lower streams");
			return DDI_FAILURE;	/* refuse to unload */
		}

		ddi_remove_minor_node(dip, NULL);

		return (DDI_SUCCESS);

	    default:
		instance = ddi_get_instance(dip);
		if (mux_pflag) 
			cmn_err(CE_CONT, 
				"mux: mux_detach(%d), i%d", cmd, instance);
		return (DDI_FAILURE);
	}
}

static int 
mux_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	int instance;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: open()");
	
	if (q->q_ptr)
		return EBUSY;
	
	if (sflag != CLONEOPEN)
		return ENXIO;

	for (instance = 0; instance < max_inst; ++ instance)
		if (!(mux_state[instance].flags & MUX_INST_INUSE))
			break;
	if (instance == max_inst)
		if (max_inst == MUX_MAX_INST)
			return EMFILE;
		else
			++max_inst;
	*devp = makedevice(getmajor(*devp), instance);

	bzero(mux_state + instance, sizeof(struct mux_state));
	mux_state[instance].flags |= MUX_INST_INUSE;
	mux_state[instance].q = q;
	WR(q)->q_ptr = q->q_ptr = (queue_t *)(mux_state + instance);

	qprocson(q);

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux open: instance %d%s",
			instance, sflag == CLONEOPEN ? " CLONE" : "");

	return 0;
}

static int 
mux_close(queue_t *q, int oflag, cred_t *credp)
{
	struct mux_state *ms;

	ms = (struct mux_state *)q->q_ptr;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux: close(), i %d", ms - mux_state);

	qprocsoff(q);

	ms->flags &= ~MUX_INST_INUSE;
	ms->q = NULL;
	ms->dip = NULL;

	return 0;
}

static int 
mux_uwput(queue_t *q, mblk_t *mp)
{
	struct mux_ctl *ctl;
	struct mux_state *ls = (struct mux_state *)q->q_ptr;
	struct mux_statistics *s;
	mblk_t *mp2;

	switch (mp->b_datap->db_type) {
	    case M_PCPROTO:
		/* high priority message: get status, non blocking */
		ctl = is_mux_ctl(mp);
		if (!ctl) {
			freemsg(mp);
			return 0;
		}
		if (ctl->cmd == MUX_C_ROUTE) {
			route_downstream(q, mp, 0);
			return 0;
		}
		else if (ctl->cmd != MUX_C_STAT) {
			freemsg(mp);
			return 0;
		}

		/* handle MUX_C_STAT */
		if (!(mp2 = mp->b_cont) || mp2->b_wptr - mp2->b_rptr != 
						sizeof(struct mux_statistics)) {
			freemsg(mp);
			return 0;
		}
		ctl->error = 0;
		s = (struct mux_statistics *)mp2->b_rptr;
		bcopy(&ls->stat, s, sizeof(struct mux_statistics));
		qreply(q, mp);
		break;

	    case M_DATA:
		freemsg(mp);
		break;

	    case M_PROTO:
		ctl = is_mux_ctl(mp);
		if (!ctl) {
			freemsg(mp);
			return 0;
		}
		route_downstream(q, mp, 0);
		break;
	
	    case M_IOCTL:
		if (q->q_first)
			putq(q, mp);
		else
			do_ioctl(q, mp);
		break;

	    case M_FLUSH:
		/* canonical flush handling, read q is unused */
		if (*mp->b_rptr & FLUSHW)
			/* flush write q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		if (*mp->b_rptr & FLUSHR) {
			/* send request back upstream without FLUSHW */
			*mp->b_rptr &= ~FLUSHW;
			qreply(q, mp);
		}
		else
			freemsg(mp); /* free message */
		break;
	
	    default:
		freemsg(mp);
		break;
	}

	return 0;
}

static int 
mux_uwsrv(queue_t *q)
{
	mblk_t *mp;

	while (mp = getq(q)) {

		switch (mp->b_datap->db_type) {
	    	    case M_IOCTL:
			do_ioctl(q, mp);
			break;

		    default:
			if (!route_downstream(q, mp, 1)) {
				/* this q will not be scheduled any more 
			 	* until backenabled by mux_lwsrv()
			 	*/
				return 0;
			}
			break;
		}
	}

	return 0;
}

static int 
mux_ursrv(queue_t *q)
{
	int n;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux_ursrv: backenabling lower read q's");

	/* Called when q is backenabled. Backenable lower read q's. */
	for (n = 0; n < max_lower; ++n)
		if (mux_lower[n].flags & MUX_LOWER_INUSE)
			qenable(RD(mux_lower[n].q));

	return 0;
}

static int 
mux_lrsrv(queue_t *q)
{
	mblk_t *mp;

	while (mp = getq(q)) {

		if (!route_upstream(q, mp, 1)) {
			/* this q will not be scheduled any more 
		 	* until backenabled by mux_ursrv()
		 	*/
			return 0;
		}
	}

	return 0;
}

static int 
mux_lrput(queue_t *q, mblk_t *mp)
{
	struct mux_ctl *ctl;
	
	ctl = is_mux_ctl(mp);
	if (!ctl || ctl->cmd != MUX_C_ROUTE) {
		freemsg(mp);
		return 0;
	}

	route_upstream(q, mp, 0);

	return 0;
}

static int 
mux_lwsrv(queue_t *q)
{
	int n;

	if (mux_pflag) 
		cmn_err(CE_CONT, "mux_lwsrv: backenabling upper write q's");

	/* Called when q is backenabled. Backenable upper write q's. */
	for (n = 0; n < max_inst; ++n)
		if (mux_state[n].flags & MUX_INST_INUSE)
			qenable(WR(mux_state[n].q));

	return 0;
}


/*
 * print a message for debug purposes
 */
static void 
print_msg(char *s, mblk_t *mp)
{
	unsigned char b[80], *p, *p2;

	strncpy((char *)b, s, sizeof(b) - 1);
	b[sizeof(b) - 1] = 0;
	for (p = b + strlen((char *)b), p2 = mp->b_rptr; 
			p < b + sizeof(b) - 1 && p2 < mp->b_wptr; )
		*p++ = *p2++;
	*p = 0;

	printf("%s", b);
}

/* Do ioctls. Ioctls are either recognized or nack'ed.
 * They are not routed downstream.
 */
static void 
do_ioctl(queue_t *q, mblk_t *mp)
{
	struct mux_state *ls = (struct mux_state *)q->q_ptr;
	struct iocblk *ioc;
	struct mux_statistics *s;
	struct linkblk *lblk;
	mblk_t *mp2;
	int n, err;

	ioc = (struct iocblk *) mp->b_rptr;
	if (ioc->ioc_count == TRANSPARENT) {
		err = EINVAL;
		goto ioc_nak;
	}
	switch (ioc->ioc_cmd) {
	    case MUX_I_GETSTAT:
		if (ioc->ioc_count != sizeof(struct mux_statistics) ||
				!(mp2 = mp->b_cont) || 
		  		mp2->b_wptr - mp2->b_rptr != 
					sizeof(struct mux_statistics)) {
			err = EINVAL;
			goto ioc_nak;
		}
		s = (struct mux_statistics *)mp2->b_rptr;
		bcopy(&ls->stat, s, sizeof(struct mux_statistics));
		break;

	    case I_LINK:
	    case I_PLINK:
	    case I_UNLINK:
	    case I_PUNLINK:
		if (ioc->ioc_count != sizeof(struct linkblk) ||
				!(mp2 = mp->b_cont) || 
		  		mp2->b_wptr - mp2->b_rptr != 
					sizeof(struct linkblk)) {
			err = EINVAL;
			goto ioc_nak;
		}
		lblk = (struct linkblk *)mp2->b_rptr;

		if (ioc->ioc_cmd == I_LINK || ioc->ioc_cmd == I_PLINK) {
			for (n = 0; n < max_lower; ++n)
				if (!(mux_lower[n].flags & MUX_LOWER_INUSE))
					break;
			if (n == max_lower) {
				if (max_lower >= MAX_MUX_LOWER) {
					err = EMFILE;
					goto ioc_nak;
				}
				++max_lower;
			}
			mux_lower[n].flags |= MUX_LOWER_INUSE;
			mux_lower[n].q = lblk->l_qbot;
			/* ioctl(fd, I_UNLINK, 0) -> EINVAL !! */
			lblk->l_index = n + 1;

			if (mux_pflag) 
				cmn_err(CE_CONT, 
					"mux ioctl: linked lower stream %d",
						lblk->l_index);
		}
		else { /* I_UNLINK */
			n = lblk->l_index - 1;
			if (n < 0 || n >= max_lower) {
				err = EINVAL;
				goto ioc_nak;
			}
			if (!(mux_lower[n].flags & MUX_LOWER_INUSE)) {
				err = ENXIO;
				goto ioc_nak;
			}

			mux_lower[n].flags &= ~MUX_LOWER_INUSE;
			mux_lower[n].q = NULL;
			while (max_lower && n == max_lower - 1 && 
				    !(mux_lower[n].flags & MUX_LOWER_INUSE)) {
				--max_lower;
				--n;
			}

			if (mux_pflag) 
				cmn_err(CE_CONT, 
					"mux ioctl: unlinked lower stream %d",
						lblk->l_index);
		}
		break;

	    default:
		err = EINVAL;
		goto ioc_nak;
	}

	/* send ACK */
	mp->b_datap->db_type = M_IOCACK;
	ioc->ioc_error = ioc->ioc_rval = 0;
	qreply(q, mp);

	return;

ioc_nak:
	mp->b_datap->db_type = M_IOCNAK;
	ioc->ioc_error = err;
	qreply(q, mp);
}

/* 
 * Check if mp is a mux_ctl (MUX_C_ROUTE or MUX_C_STAT) message.
 */
static struct mux_ctl *
is_mux_ctl(mblk_t *mp)
{
	struct mux_ctl *ctl;
	unsigned ty;

	ty = mp->b_datap->db_type;
	if (ty != M_PCPROTO && ty != M_PROTO)
		return NULL;

	if (mp->b_wptr - mp->b_rptr != sizeof(struct mux_ctl))
		return NULL;

	ctl = (struct mux_ctl *)mp->b_rptr;
	if (ctl->cmd != MUX_C_ROUTE && ctl->cmd != MUX_C_STAT)
		return NULL;
	
	return ctl;
}

/*
 * Route message downstream to lower mux half. Returns true if
 * message was routed or discarded.
 */
int
route_downstream(queue_t *q, mblk_t *mp, int called_by_srv)
{
	struct mux_ctl *ctl = is_mux_ctl(mp);
	queue_t *lq;
	int i;

	if (!ctl) {
		cmn_err(CE_WARN, "mux: internal error in route_downstream");
		freemsg(mp);
		return 1;
	}

	ctl->error = 0;
	i = ctl->down_idx - 1;
	if (i < 0 || i >= max_lower || !(mux_lower[i].flags & MUX_LOWER_INUSE)){
		/* error, send back */
		ctl->error = MUX_ERR_BAD_L_INDEX;
		qreply(q, mp);
		cmn_err(CE_WARN, "mux route_downstream: bad l_index %d", i);
		return 1;
	}
	lq = mux_lower[i].q;

	if (called_by_srv) {
		if (canputnext(lq))
			putnext(lq, mp);
		else {
			putbq(q, mp);
			return 0;
		}
	}
	else if (!q->q_first && canputnext(lq))
		putnext(lq, mp);
	else {
		putq(q, mp);
		return 0;
	}
	
	return 1;
}

/*
 * Route message upstream to upper mux half. Returns true if
 * message was routed or discarded.
 */
int
route_upstream(queue_t *q, mblk_t *mp, int called_by_srv)
{
	struct mux_ctl *ctl = is_mux_ctl(mp);
	queue_t *uq;
	int i;

	if (!ctl) {
		cmn_err(CE_WARN, "mux: internal error in route_upstream");
		freemsg(mp);
		return 1;
	}

	ctl->error = 0;
	i = ctl->up_idx;
	if (i < 0 || i >= max_inst || !(mux_state[i].flags & MUX_INST_INUSE)){
		cmn_err(CE_WARN, "mux route_upstream: bad minor %d", i);
		freemsg(mp);
		return 1;
	}
	uq = mux_state[i].q;

	if (called_by_srv) {
		if (canputnext(uq))
			putnext(uq, mp);
		else {
			putbq(q, mp);
			return 0;
		}
	}
	else if (!q->q_first && canputnext(uq))
		putnext(uq, mp);
	else {
		putq(q, mp);
		return 0;
	}
	
	return 1;
}
